package org.hybird.scenile.shapes.converters;

import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/*
 * Author: Leon Franzen
 * Source: http://mail-archives.apache.org/mod_mbox/xmlgraphics-batik-users/200408.mbox/<83E753BE7B6A324ABB336245BAF1DAAC01A8BD07@mailserver.metatv-ds.metatv.com>
 * 
 * Slightly modified by Rmy Rakic
 */

public class SVGPathParser
{
	private static final Pattern elemPtn = Pattern.compile ("([a-zA-Z])\\s*([^a-zA-Z]*)?");
	private static final Pattern coordPtn = Pattern.compile ("\\s*([^,\\s]+)\\s*[\\s,]?\\s*([^,\\s]+)");

	public static GeneralPath parse (String string)
	{
		GeneralPath gp = new GeneralPath ();
		SVGPathParser parser = new SVGPathParser ();
		try
		{
			parser.fillPath (gp, string);
		}
		catch (Exception e)
		{
			e.printStackTrace();
			throw new RuntimeException (e);
		}
		return gp;
	}

	/**
	 * @param gp
	 * @param string
	 * @throws Exception
	 */
	private void fillPath (GeneralPath gp, String string) throws Exception
	{
		Matcher m = elemPtn.matcher (string);
		while (m.find ())
		{
			// System.out.println(">" + m.group(1));
			switch (m.group (1).charAt (0))
			{
				case ('M'):
				case ('m'):
					handleMoveTo (gp, m.group (2));
					break;
				case ('L'):
				case ('l'):
					handleLineTo (gp, m.group (2));
					break;
	
				case ('C'):
				case ('c'):
					handleCurveTo (gp, m.group (2));
					break;
	
				case ('Q'):
				case ('q'):
					handleQuadTo (gp, m.group (2));
					break;
	
				case ('Z'):
				case ('z'):
					gp.closePath ();
					break;
	
				default:
				// do Nothing
			}
		}
	}

	/**
	 * @param gp
	 * @param string
	 * @throws Exception
	 */
	private void handleQuadTo (GeneralPath gp, String string) throws Exception
	{
		List<Point2D.Float> v = parseCoords (string);
		Iterator<Point2D.Float> it = v.iterator ();
		while (it.hasNext ())
		{
			Point2D.Float p1 = it.next ();
			Point2D.Float p2 = it.next ();

			if (p1 == null || p2 == null)
			{
				throw new Exception ("");
			}
			gp.quadTo (p1.x, p1.y, p2.x, p2.y);
		}

	}

	/**
	 * @param gp
	 * @param string
	 * @throws Exception
	 */
	private void handleCurveTo (GeneralPath gp, String string) throws Exception
	{
		List<Point2D.Float> v = parseCoords (string);
		Iterator<Point2D.Float> it = v.iterator ();
		while (it.hasNext ())
		{
			Point2D.Float p1 = it.next ();
			Point2D.Float p2 = it.next ();
			Point2D.Float p3 = it.next ();

			if (p1 == null || p2 == null || p3 == null)
			{
				throw new Exception ("");
			}
			gp.curveTo (p1.x, p1.y, p2.x, p2.y, p3.x, p3.y);
		}

	}

	/**
	 * @param gp
	 * @param string
	 * @throws Exception
	 */
	private void handleLineTo (GeneralPath gp, String string) throws Exception
	{
		List<Point2D.Float> v = parseCoords (string);
		handleLineTo (gp, v);
	}

	private void handleLineTo (GeneralPath gp, List<Point2D.Float> v) throws Exception
	{
		Iterator<Point2D.Float> it = v.iterator ();
		while (it.hasNext ())
		{
			Point2D.Float p1 = it.next ();
			if (p1 == null)
			{
				throw new Exception ("");
			}
			gp.lineTo (p1.x, p1.y);
		}
	}

	/**
	 * @param gp
	 * @param string
	 * @throws Exception
	 */
	private void handleMoveTo (GeneralPath gp, String string) throws Exception
	{
		List<Point2D.Float> coords = parseCoords (string);
		Iterator<Point2D.Float> it = coords.iterator ();
		Point2D.Float mpt = it.next ();
		if (mpt == null)
		{
			throw new Exception ("");
		}
		gp.moveTo (mpt.x, mpt.y);
		handleLineTo (gp, coords);
	}

	/**
	 * @param string
	 * @return
	 */
	private List<Point2D.Float> parseCoords (String string)
	{
		string = string.trim ();
		String coords[] = string.split ("\\s+");
		List<Point2D.Float> v = new ArrayList<Point2D.Float>();
		for (int i = 0; i < coords.length; i++)
		{
			Matcher m = coordPtn.matcher (coords[i]);
			if (m.find ())
			{
				// System.out.println(" (" + m.group(1) + " , " + m.group(2) + ")");
				Point2D.Float pt = new Point2D.Float (Float.parseFloat (m.group (1)), Float.parseFloat (m.group (2)));
				v.add (pt);
			}
		}
		return v;
	}
}