package org.hybird.scenile.util;

/*
*
* Copyright 2007 Sun Microsystems, Inc.  All Rights Reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   - Redistributions of source code must retain the above copyright
*     notice, this list of conditions and the following disclaimer.
*
*   - Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the
*     documentation and/or other materials provided with the distribution.
*
*   - Neither the name of Sun Microsystems nor the names of its
*     contributors may be used to endorse or promote products derived
*     from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
* IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
* THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
* PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsEnvironment;
import java.awt.Image;
import java.awt.Paint;
import java.awt.Point;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.Transparency;
import java.awt.event.ActionEvent;
import java.awt.event.HierarchyEvent;
import java.awt.event.HierarchyListener;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.awt.image.VolatileImage;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.Timer;
import javax.swing.ToolTipManager;
import javax.swing.UIManager;
import javax.swing.border.EmptyBorder;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeExpansionListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;

import com.sun.scenario.scenegraph.SGClip;
import com.sun.scenario.scenegraph.SGComponent;
import com.sun.scenario.scenegraph.SGComposite;
import com.sun.scenario.scenegraph.SGGroup;
import com.sun.scenario.scenegraph.SGImage;
import com.sun.scenario.scenegraph.SGNode;
import com.sun.scenario.scenegraph.SGParent;
import com.sun.scenario.scenegraph.SGShape;
import com.sun.scenario.scenegraph.SGText;
import com.sun.scenario.scenegraph.SGTransform;
import com.sun.scenario.scenegraph.fx.FXAbstractShape;
import com.sun.scenario.scenegraph.fx.FXComponent;
import com.sun.scenario.scenegraph.fx.FXImage;
import com.sun.scenario.scenegraph.fx.FXNode;
import com.sun.scenario.scenegraph.fx.FXShape;
import com.sun.scenario.scenegraph.fx.FXText;

/**
 * Imported from the scenegraph-demos.dev.java.net workspace
 * --------------------------------------------------------------------------
 * 
 * Component which shows a scrollable, hierarchical display of the nodes in a
 * Scenario scene graph.
 *
 * Static convenience methods are provided to make it easy to display the
 * tree view in a dialog.
 *
 * <pre><code>
 * Example usage:
 *
 *     // create scene graph ...
 *     SGGroup topNode = new SGGroup();
 *     JSGPanel scenePanel = new JSGPanel();
 *     scenePanel.setScene(topNode);
 *
 *     <b>// Install hotkey (defaults to control-T) for bringing up tree view in a dialog;
 *     // since refresh interval is 0, dialog will have a Refresh button.
 *     SceneTreeView view = SceneTreeView.installHotkeyDisplay(scenePanel, topNode, 0);</b>
 *
 * </code></pre>
 * @author aim
 */
public class SceneTreeView extends JPanel {
    
    private static final int DEFAULT_ROW_HEIGHT = 25;
    
    private static final GraphicsConfiguration CONFIGURATION =
            GraphicsEnvironment.getLocalGraphicsEnvironment().
                    getDefaultScreenDevice().getDefaultConfiguration();
    
    private static VisibilityListener visibilityListener; // shared by all SceneTreeView instances
    
    /**
     * Installs a hotkey accelerator (&quot;control-T&quot;) to bring up the tree view
     * on the scene graph.  If the refresh interval is <= 0, then a refresh button will be added 
     * to the dialog to allow the user to explicitly refresh the tree view when the scene graph changes.
     *
     * @param origin the component where the hotkey is installed
     * @param scene the root node in the scene graph
     * @param refreshInterval the amount of time (in milliseconds) between automatic refreshes of
     *        of scene graph tree view; if value is <= 0, then automatic refresh is disabled.
     */
    public static SceneTreeView installHotkeyDisplay(JComponent origin, SGNode scene, int refreshInterval) {
        return installHotkeyDisplay(origin, scene, refreshInterval, KeyStroke.getKeyStroke("ctrl T"));
    }

    /**
     * Installs the specified hotkey accelerator to bring up the tree view on the scene graph.
     * If the refresh interval is <= 0, then a refresh button will be added to the dialog to 
     * allow the user to explicitly refresh the tree view when the scene graph changes.
     *
     * @param origin the component where the hotkey is installed
     * @param scene the root node in the scene graph
     * @param refreshInterval the amount of time (in milliseconds) between automatic refreshes of
     *        of scene graph tree view; if value is <= 0, then automatic refresh is disabled.
     * @param hotkey the key stroke to be used to trigger the display action
     */    
    public static SceneTreeView installHotkeyDisplay(JComponent origin, SGNode scene, int refreshInterval,
            KeyStroke hotkey) {
        SceneTreeView sceneTreeView = new SceneTreeView();
        sceneTreeView.setRefreshInterval(refreshInterval);
        Action displayAction = new DisplayAction(sceneTreeView, scene);
        origin.getActionMap().put("DisplayTreeView", displayAction);
        origin.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(hotkey, "DisplayTreeView");
        return sceneTreeView;
    }
    
    /**
     * Creates a dialog containing a tree view for the specified scene graph. If the refresh 
     * interval is <= 0, then a refresh button will be added to the dialog to allow the
     * user to explicitly refresh the tree view when the scene graph changes.
     *
     * @param origin the component used to determine the toplevel container for which the dialog
     *        should be placed relative to
     * @param scene the root node in the scene graph
     * @param refreshInterval the amount of time (in milliseconds) between automatic refreshes of
     *        of scene graph tree view; if value is <= 0, then automatic refresh is disabled.
     */
    public static SceneTreeView showSceneGraphDialog(Component origin, SGNode scene, int refreshInterval) {
        SceneTreeView sceneTreeView = new SceneTreeView(scene);
        sceneTreeView.setRefreshInterval(refreshInterval);
        JDialog dialog = createSceneGraphDialog(origin, sceneTreeView);
        dialog.setVisible(true);
        
        return sceneTreeView;        
    }

    /**
     * Creates a dialog containing the specified tree view.  If the refresh interval on
     * the tree view is <= 0, then a refresh button will be added to the dialog to allow the
     * user to explicitly refresh the tree view when the scene graph changes.
     *
     * @param origin the component used to determine the toplevel container for which the dialog
     *        should be placed relative to
     * @param root the root node in the scene graph
     * @param refreshInterval the amount of time (in milliseconds) between automatic refreshes of
     *        of scene graph tree view; if value is <= 0, then automatic refresh is disabled.
     */
    public static JDialog createSceneGraphDialog(Component origin, SceneTreeView sceneTreeView) {
        
        JDialog dialog = new JDialog();     
        dialog.setTitle("Scene Graph TreeView");
        dialog.add(sceneTreeView);
        
        JPanel panel = new JPanel();
        dialog.getContentPane().add(panel, BorderLayout.SOUTH);
        
        if (sceneTreeView.getRefreshInterval() <= 0) {
            JButton button = new JButton(sceneTreeView.getActionMap().get("RefreshTreeView"));
            button.setText("Refresh");
            panel.add(button);
        }
        dialog.pack();
        if (origin != null && origin.isDisplayable()) {
            Point p = new Point(origin.getWidth(), 0);
            SwingUtilities.convertPointToScreen(p, origin);
            dialog.setLocation(p);
        }
        return dialog;
    }

    /**
     * (Courtesy of Filthy Rich Client examples)
     * <p>Returns a new compatible image of the specified width and height, and
     * the same transparency setting as the image specified as a parameter.</p>
     *
     * @see java.awt.Transparency
     * @see #createCompatibleImage(java.awt.image.BufferedImage)
     * @see #createCompatibleImage(int, int)
     * @see #createTranslucentCompatibleImage(int, int)
     * @see #loadCompatibleImage(java.net.URL)
     * @see #toCompatibleImage(java.awt.image.BufferedImage)
     * @param width the width of the new image
     * @param height the height of the new image
     * @param image the reference image from which the transparency of the new
     *   image is obtained
     * @return a new compatible <code>BufferedImage</code> with the same
     *   transparency as <code>image</code> and the specified dimension
     */
    protected static BufferedImage createCompatibleImage(Image image,
                                                      int width, int height) {
        int translucency = Transparency.TRANSLUCENT;
        if (image instanceof BufferedImage) {
            translucency = ((BufferedImage) image).getTransparency();
        } else if (image instanceof VolatileImage) {
            translucency = ((VolatileImage) image).getTransparency();
        }
        return CONFIGURATION.createCompatibleImage(width, height,
                                                   translucency);
    }

    
    /**
     * (Courtesy of Filthy Rich Clients examples)
     * <p>Returns a thumbnail of a source image.</p>
     * <p>This method favors speed over quality. When the new size is less than
     * half the longest dimension of the source image,
     * {@link #createThumbnail(BufferedImage, int)} or
     * {@link #createThumbnail(BufferedImage, int, int)} should be used instead
     * to ensure the quality of the result without sacrificing too much
     * performance.</p>
     *
     * @see #createThumbnailFast(java.awt.image.BufferedImage, int)
     * @see #createThumbnail(java.awt.image.BufferedImage, int)
     * @see #createThumbnail(java.awt.image.BufferedImage, int, int)
     * @param image the source image
     * @param newWidth the width of the thumbnail
     * @param newHeight the height of the thumbnail
     * @return a new compatible <code>BufferedImage</code> containing a
     *   thumbnail of <code>image</code>
     * @throws IllegalArgumentException if <code>newWidth</code> is larger than
     *   the width of <code>image</code> or if code>newHeight</code> is larger
     *   than the height of <code>image</code> or if one of the dimensions
     *   is &lt;= 0
     */
    private static BufferedImage createThumbnailFast(Image image,
                                                     int newWidth,
                                                     int newHeight)
    {
        if (newWidth < 1) newWidth = 1;
        if (newHeight < 1) newHeight = 1;

        BufferedImage temp = createCompatibleImage(image, newWidth, newHeight);
        Graphics2D g2 = temp.createGraphics();
        g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION,
                            RenderingHints.VALUE_INTERPOLATION_BILINEAR);
        g2.drawImage(image, 0, 0, newWidth, newHeight, null);
        g2.dispose();

        return temp;
    }

    
    private DefaultTreeModel treeModel; 
    private DefaultMutableTreeNode treeModelRoot;
    
    private JTree tree;
    
    private SGNode scene;
    
    private Timer timer;
    private int refreshInterval = 0;
    private boolean configuringExpansion = false;
    private boolean collapseFXNodes = true;

    /**
     * Creates a new instance of SceneTreeView
     */
    public SceneTreeView() {
        setLayout(new BorderLayout());
        tree = new JTree();
        tree.setShowsRootHandles(true);
        tree.setExpandsSelectedPaths(true);
        tree.setCellRenderer(new SGNodeRenderer());
        tree.addTreeExpansionListener(new NodeExpansionListener());
        tree.setVisibleRowCount(20);
        setRowHeight(DEFAULT_ROW_HEIGHT);
        ToolTipManager.sharedInstance().registerComponent(tree);
      
        add(new JScrollPane(tree), BorderLayout.CENTER);
        
        getActionMap().put("RefreshTreeView", new RefreshAction());
    }
    
    public SceneTreeView(SGNode root) {
        this();
        setScene(root);
    }
    
    /**
     * Sets the refresh interval (in milliseconds) for this tree view. 
     * Since the scene graph does not emit events when it changes, the only way
     * to keep the tree view current as the scene graph mutates is to refresh the view
     * periodically.  The default is 0, which means automatic refresh is disabled and
     * the tree view will not update to reflect changes in the scene graph.
     *
     * @param interval the amount of time (in milliseconds) between automatic refreshes of
     *        of scene graph tree view; if value is <= 0, then automatic refresh is disabled.
     */
    public void setRefreshInterval(int interval) {
        int oldInterval = this.refreshInterval;
        this.refreshInterval = interval;
        
        if (interval > 0) {
            if (oldInterval <= 0) {
                if (timer == null) {
                    timer = new Timer(interval, getActionMap().get("RefreshTreeView"));
                    timer.setRepeats(true);
                    if (visibilityListener == null) {
                        visibilityListener = new VisibilityListener();
                    }
                }
                // Only run refresh is tree view is showing on screen
                addHierarchyListener(visibilityListener);
                if (isShowing()) {
                    timer.start();
                }
            } else {
                // just changing the delay
                timer.setDelay(interval);
            }
            
        } else if (oldInterval > 0) {
            // turn off automatic refresh
            timer.stop();
            removeHierarchyListener(visibilityListener);
        } 
        firePropertyChange("refreshInterval", oldInterval, interval);
    }
    
    public int getRefreshInterval() {
        return refreshInterval;
    }

    public void setDefaultFXNodeCollapseState(boolean collapsed) {
        collapseFXNodes = collapsed;
    }

    public boolean getDefaultFXNodeCollapseState() {
        return collapseFXNodes;
    }

    public void setRowHeight(int rowHeight) {
        int oldRowHeight = tree.getRowHeight();
        tree.setRowHeight(rowHeight);
        firePropertyChange("rowHeight", oldRowHeight, rowHeight);
    }
    
    public int getRowHeight() {
        return tree.getRowHeight();
    }
    
    public void setScene(SGNode scene) {
        SGNode oldScene = this.scene;
        this.scene = scene;
        
        treeModelRoot = createNodeBranch(scene);
        if (treeModel == null) {
            treeModel = new DefaultTreeModel(treeModelRoot);
            tree.setModel(treeModel);
        } else {
            treeModel.setRoot(treeModelRoot);
        }
        treeModel.nodeStructureChanged(treeModelRoot);
        ((SGNodeRenderer)tree.getCellRenderer()).reset();
        
        configuringExpansion = true;
        configureExpandedState(new TreePath(treeModelRoot));
        configuringExpansion = false;
        
        firePropertyChange("scene", oldScene, scene);
    }
    
    public SGNode getScene() {
        return scene;
    }
    
    public void refreshScene() {
        SGNode selectedNode = null;
        DefaultMutableTreeNode treeNode = (DefaultMutableTreeNode)tree.getLastSelectedPathComponent();
        if (treeNode != null) {
            selectedNode = getSGNode(treeNode);
        }
        treeModelRoot = createNodeBranch(scene);
       
        if (treeModel == null) {
            treeModel = new DefaultTreeModel(treeModelRoot);
            tree.setModel(treeModel);
        } else {
            treeModel.setRoot(treeModelRoot);
        }
        
        treeModel.nodeStructureChanged(treeModelRoot);
        ((SGNodeRenderer)tree.getCellRenderer()).reset();
        
        configuringExpansion = true;
        configureExpandedState(new TreePath(treeModelRoot));
        configuringExpansion = false;
        
        if (selectedNode != null) {
            TreePath path = getPathForSGNode(selectedNode);
            if (path != null) {
                tree.setSelectionPath(getPathForSGNode(selectedNode));
            }
        }
    }
    
    public DefaultMutableTreeNode createNodeBranch(SGNode node) {
        DefaultMutableTreeNode treeNode = new DefaultMutableTreeNode(node);

        if (node instanceof SGParent) {
            List<SGNode> children = ((SGParent)node).getChildren();
            for(SGNode child: children) {
                treeNode.add(createNodeBranch(child));
            }
            // initialize expanded/collapsed state of any new nodes
            Boolean collapsed = getNodeCollapsed(node);
            if (collapsed == null) {
                // new node
                setNodeCollapsed(node, node instanceof FXNode? collapseFXNodes : false);
            }
        } 
        return treeNode;
    }
        
    protected void configureExpandedState(TreePath treePath) {
        TreeNode treeNode = (TreeNode)treePath.getLastPathComponent();

        if (!treeNode.isLeaf()) {
            SGNode node = getSGNode(treeNode);
            // Traverse children first
            if (treeNode.getChildCount() >= 0) {
                Enumeration<?> e = treeNode.children();
                while (e.hasMoreElements()) {
                    TreeNode n = (TreeNode)e.nextElement();
                    TreePath path = treePath.pathByAddingChild(n);
                    configureExpandedState(path);
                }
            }
            if (isNodeCollapsed(node)) {
                tree.collapsePath(treePath);
            } else {
                tree.expandPath(treePath);
            }
        }

    } 
    
    protected SGNode getSGNode(Object treeNode) {
        Object o = ((DefaultMutableTreeNode)treeNode).getUserObject();
        if (!(o instanceof SGNode)) {
            // REMIND: Why is this happening?
            SGGroup g = new SGGroup();
            g.setID("Unknown: "+o);
            o = g;
        }
        return (SGNode) o;
    }
    
    protected TreePath getPathForSGNode(SGNode node) {
        Enumeration<?> nodes = treeModelRoot.breadthFirstEnumeration();
        while(nodes.hasMoreElements()) {
            DefaultMutableTreeNode treeNode = (DefaultMutableTreeNode)nodes.nextElement();
            if (getSGNode(treeNode).equals(node)) {
               return new TreePath(treeModel.getPathToRoot(treeNode));
            }
        }
        return null; // node not found        
    }
    
    protected Boolean getNodeCollapsed(SGNode node) {
        return (Boolean)node.getAttribute("SceneTreeView.collapsed");
    }
    
    protected boolean isNodeCollapsed(SGNode node) {
        Boolean collapsed = getNodeCollapsed(node);
        if (collapsed != null) {
            return collapsed.booleanValue();
        }
        return false; // not set
    }
    
    protected void setNodeCollapsed(SGNode node, boolean collapsed) {
        node.putAttribute("SceneTreeView.collapsed", Boolean.valueOf(collapsed));
    }
    
    // listener to track when nodes are collapsed so when the tree view is refreshed
    // with an updated scene graph, we can restore the collapsed/expanded states of
    // each node
    private class NodeExpansionListener implements TreeExpansionListener {
        public void treeExpanded(TreeExpansionEvent event) {
            if (!configuringExpansion) { 
                SGNode node = getSGNode(event.getPath().getLastPathComponent());
                setNodeCollapsed(node, false);
            }
        }
        public void treeCollapsed(TreeExpansionEvent event) {
            if (!configuringExpansion) {                
                SGNode node = getSGNode(event.getPath().getLastPathComponent());
                setNodeCollapsed(node, true);
            }
        }
    }
    
    public class SGNodeRenderer extends DefaultTreeCellRenderer {
        private Color defaultColor = new Color(240, 240, 250);
        
        private HashMap<Class<?>,Color> colorMap;
        private HashMap<SGNode,ImageIcon> thumbnailCache;
        private SGNode node;
        private boolean nodeIsSelected = false;
        private boolean nodeHasFocus = false;
        
        public SGNodeRenderer() {
            super();
            setBorder(new EmptyBorder(2,2,2,2));
            setHorizontalTextPosition(JLabel.LEFT);
            
            colorMap = new HashMap<Class<?>, Color>();
            colorMap.put(SGShape.class, new Color(255, 228, 206));
            colorMap.put(FXShape.class, new Color(245, 208, 186));
            colorMap.put(SGText.class, new Color(199, 255, 185));
            colorMap.put(FXText.class, new Color(179, 245, 165));
            colorMap.put(SGComponent.class, new Color(210, 210, 255));
            colorMap.put(FXComponent.class, new Color(190, 190, 245));
            colorMap.put(SGImage.class, new Color(255, 185, 185));
            colorMap.put(FXImage.class, new Color(245, 165, 165));
            colorMap.put(SGComposite.class, new Color(215, 215, 225));
            colorMap.put(SGTransform.class, new Color(220, 220, 230));
            colorMap.put(SGClip.class, new Color(225, 225, 235));
            colorMap.put(FXAbstractShape.class, new Color(255, 226, 185));
            colorMap.put(FXNode.class, new Color(202, 175, 255));
            
            thumbnailCache = new HashMap<SGNode,ImageIcon>();
            
        }
        
        public void reset() {
            node = null;
            thumbnailCache.clear();
        }
        
        @Override
        public Component getTreeCellRendererComponent(
                JTree tree,
                Object value,
                boolean isSelected,
                boolean isExpanded,               
                boolean isLeaf,
                int row,
                boolean hasFocus) {
            super.getTreeCellRendererComponent(tree, value, false, isExpanded, isLeaf, row, false);
            
            DefaultMutableTreeNode treeNode = (DefaultMutableTreeNode)value;
            node = getSGNode(treeNode);
            nodeIsSelected = isSelected;
            nodeHasFocus = hasFocus;
            String className = node.getClass().getSimpleName();
            String id = node.getID();
            Rectangle2D bounds = node.getBounds();
            
            Icon icon = null;
            
            String text = "<html><b>" + className + "</b><font color=\"#0000EE\">" +
                    (id != null && !id.equals("")? " \"" + id + "\"" : "") +
                    "</font>";
                    
            String tooltipText = "<html>" +
                    "bounds=["+bounds.getX() + "," + bounds.getY() + " " + 
                          bounds.getWidth()+"x"+bounds.getHeight() + "]" + (nodeHasFocus ? " [focused]" : "");
                    
            if (node instanceof SGComposite) {
                text += "<font color=\"#0C0C0C\" size=\"3\">" +
                        String.format(" (%2.2f",((SGComposite)node).getOpacity()) + ") </font>";
                
            } else if (node instanceof SGComponent) {
                Component component = ((SGComponent)node).getComponent();
                text += " (" + (component != null? 
                    component.getClass().getSimpleName() : "null") + ")";
                
            } else if (node instanceof FXComponent) {
                Component component = ((FXComponent)node).getComponent();
                text += " (" + (component != null? 
                    component.getClass().getSimpleName() : "null") + ")";
                
            } else if (node instanceof SGText) {
                String tvalue = ((SGText)node).getText();
                text += " (\""+ tvalue.substring(0, Math.min(tvalue.length(),12)) + 
                                 (tvalue.length() > 12? "...\"" : "\")");
                
            } else if (node instanceof FXText) {
                String tvalue = ((FXText)node).getText();
                text += " (\""+ tvalue.substring(0, Math.min(tvalue.length(),12)) + 
                                 (tvalue.length() > 12? "...\"" : "\")");
                
            } else if (node instanceof SGImage || node instanceof FXImage)  {
                // remind(aim): thumbnails will be duplicated for FXImage; need to fix
                ImageIcon thumbnail = thumbnailCache.get(node);
                if (thumbnail == null) {
                    Image image = node instanceof SGImage? ((SGImage)node).getImage() :
                        ((FXImage)node).getImage();
                    if (image != null) {
                        float scale = DEFAULT_ROW_HEIGHT/(float)image.getHeight(null);
                        //BufferedImage thumbnailImage = (BufferedImage)image;
                        if (scale < 1.0) {
                            image = SceneTreeView.createThumbnailFast(image, 
                                (int)(image.getWidth(null) * scale), (int)(image.getHeight(null) * scale));
                        } 
                        thumbnail = new ImageIcon(image);
                        thumbnailCache.put(node, thumbnail);
                    }
                }
                icon = thumbnail;
                
                
            } else if (node instanceof SGTransform.Rotate) {
            
                text += " ("+((SGTransform.Rotate)node).getRotation() + ")";
                
            } else if (node instanceof SGTransform.Translate) {
                text += " ("+
                        ((SGTransform.Translate)node).getTranslateX() + "," +
                        ((SGTransform.Translate)node).getTranslateY() + ")";
                
            } else if (node instanceof SGTransform.Scale) {
                text += " ("+((SGTransform.Scale)node).getScaleX() +
                        "x" + ((SGTransform.Scale)node).getScaleY() + ")";
                
            } else if (node instanceof SGTransform.Shear) {
                text += " ("+((SGTransform.Shear)node).getShearX() +
                        "," + ((SGTransform.Shear)node).getShearY() + ")";
                
            } else if (node instanceof SGClip) {
                
            } else if (node instanceof FXNode) {
                
            }
                
            
            if (!node.isVisible()) {
                 text += " <i>(invisible)</i>";
            }
            setIcon(icon);
            setText(text + "</html>");
            setToolTipText(tooltipText + "</html>");
            setOpaque(false);
            return this;            
            
        }
        
        @Override

        protected void paintComponent(Graphics g) {
            Graphics2D g2 = (Graphics2D)g;
            Dimension size = getSize();
            
            Color color = colorMap.get(node.getClass());
            if (color == null) {
                color = defaultColor;
            }
            g2.setColor(color);
            g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
            g2.fillRoundRect(0, 2, size.width-1, size.height-4, 12, 12);
            

            Color borderColor = nodeIsSelected? 
            	UIManager.getColor("Tree.selectionBorderColor") != null ?
            			UIManager.getColor("Tree.selectionBorderColor").darker() :
            			color :	
                color.darker();
            Paint borderPaint = nodeIsSelected? new GradientPaint(0, 0, borderColor, 
                    size.width/2, size.height, borderColor.brighter(), true) : borderColor;
            g2.setPaint(borderPaint);
                                
            Stroke stroke = null;
            float thickness = nodeIsSelected? 3.0f : 
                (node instanceof FXNode? 2.0f : 1.0f);
            if (node instanceof SGParent) {                
                float dash[] = new float[2];
                dash[0] = 3f;
                dash[1] = 2f;
                stroke = new BasicStroke(thickness, BasicStroke.CAP_BUTT,
                        BasicStroke.JOIN_BEVEL, 1.0f, dash, 3f);
                
            } else {
                stroke = new BasicStroke(thickness);
            }
            g2.setStroke(stroke);            
            g2.drawRoundRect(0, 2, size.width-1, size.height-4, 12, 12);

            super.paintComponent(g2);
            
        }        
    } 
    
    // listener which will turn the refresh timer on/off as this component
    // is made visibile/invisible
    private static class VisibilityListener implements HierarchyListener {
        public void hierarchyChanged(HierarchyEvent event) {
            if ((event.getChangeFlags() & HierarchyEvent.SHOWING_CHANGED) > 1) {
                SceneTreeView treeview = (SceneTreeView)event.getComponent();
                if (treeview.isShowing()) {
                    if (treeview.getRefreshInterval() > 0) {
                        treeview.timer.start();
                    }
                } else {
                    if (treeview.timer.isRunning()) {
                        treeview.timer.stop();
                    }
                }
            }
        }
    }
    
    public class RefreshAction extends AbstractAction {
        
        public RefreshAction() {
            super("RefreshTreeView");
        }
        
        public void actionPerformed(ActionEvent event) {
            refreshScene();
        }
    }
    
    public static class DisplayAction extends AbstractAction {
        private SceneTreeView sceneTreeView;
        private SGNode scene;
        private JDialog dialog;
        
        public DisplayAction(SceneTreeView sceneTreeView, SGNode scene) {
            super("DisplayTreeView");
            this.sceneTreeView = sceneTreeView;
            this.scene = scene;
        }
        
        public void actionPerformed(ActionEvent event) {
            Component component = (Component)event.getSource();
            sceneTreeView.setScene(scene); // set scene lazily
            if (dialog == null) {
                dialog = SceneTreeView.createSceneGraphDialog(component, sceneTreeView);
            }
            dialog.setVisible(true);
        }
    }
}
